/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/


#include <Winsock2.h>
#include <Mswsock.h>
#include <stdio.h>

#pragma comment(lib, "ws2_32.lib")
#pragma comment(lib, "Mswsock.lib")

int nPortNum=12396;

#define MAX_DEBUG_CHAR 2048

#ifdef _DEBUG
#define ATRACELINE
#endif

#ifdef ATRACELINE
void
ATraceStr(char *str)
{
	fprintf(stdout, str);
}

void _cdecl
ATrace(char far *format, ...)
{
	char buffer[MAX_DEBUG_CHAR];
	va_list args;


	va_start(args, format);
	wvsprintf(buffer, format, args);

	buffer[MAX_DEBUG_CHAR-1] = '\0';	
	
	ATraceStr(buffer);
	
	va_end(args);
}

void _cdecl
ATraceLn(char far *format,...)
{
	char buffer[MAX_DEBUG_CHAR];
	int wLen;
	va_list args;

	va_start(args, format);
	wvsprintf(buffer, format, args);
	wLen = strlen(buffer);
	buffer[wLen++] = '\n';
	buffer[wLen++] = '\r';
	buffer[wLen] = '\0';
	
	ATraceStr(buffer);
	
	va_end(args);
}
#else
inline void ATrace(char *format,...)
{
	return;
}

inline void ATraceLn(char *format,...)
{
	return;
}

#endif

inline int myitoa(int i, char* psz)
{
	if (i<0)
	{
		i = -i;
		*psz++ = '-';
	}

	char *t = psz;
	int nc;

	do
	{
		*t++ = '0' + (i%10);
		i /= 10;
	}
	while (i);

	nc = t-psz;
	*t-- = 0;

	while (psz < t)
	{
		char c;

		c = *psz;
		*psz++ = *t;
		*t-- = c;
	}

	return nc;
}

template <class T, int SIZE> class MyStack
{
private:
	T	rgT[SIZE];
	int	top;
	
	CRITICAL_SECTION cs;
public:
	MyStack()
		:top(-1)
	{
		::InitializeCriticalSection(&this->cs);
	}

	~MyStack()
	{
		::DeleteCriticalSection(&this->cs);
	}

	int Top(void) const {return this->top;}

	bool Push(T t)
	{
		bool fErr;
		
		::EnterCriticalSection(&this->cs);
		fErr = (this->top >= SIZE-1);
		if (fErr)
		{
			ATraceLn("Error. Stack overflow!");
		}
		else
		{
			rgT[++this->top] = t;
		}
		::LeaveCriticalSection(&this->cs);

		return fErr;
	}

	bool Pop(T* pt)
	{
		bool fErr;

		::EnterCriticalSection(&this->cs);
		fErr = (this->top<0);
		if (fErr)
		{
			ATraceLn("Error. Stack underflow!");
		}
		else
		{
			*pt = rgT[this->top--];
		}
		::LeaveCriticalSection(&this->cs);

		return fErr;
	}
};


#define MAX_BUFFER_SIZE (100*1024)
#define MAX_RESOURCE_BUFFER_SIZE (2*1024)
#define MAX_SOCKET_POOL	5000
#define MAX_OBJ_POOL	5000


long gSocketCounter=0;
long nSocketCount=0;


class cHTTPAction
{
private:
	class cSocket *pSock;

public:
	cHTTPAction(class cSocket *p = 0)
		:pSock(p)
	{

	}

	virtual Write(void)
	{

	}
};

class cSocket
{
private:
	typedef enum
	{
		eioshBeginHeader,
		eioshReadCommand,
		eioshReadResource,
		eioshReadVersion,
		eioshBeginLine,
		eioshCR1,
		eioshLF1,
		eioshCR2,
		eioshLF2,
	}
	EnumIOHeaderState;

public:
	typedef enum
	{
		etsInit,
		etsRead,
		etsWrite,
		etsWriteAndDisconnect,
		etsDead
	}
	EnumTransmissionState;

	typedef struct
	{
		OVERLAPPED ol;
		cSocket *pSocket;
		EnumTransmissionState ets;
	} OverlappedEx;

private:
	SOCKADDR_IN sockClient;
	EnumIOHeaderState eiohState;

	int nLenCommand;
	char szCommand[20];
	int nLenResource;
	char szResource[MAX_RESOURCE_BUFFER_SIZE];

	HANDLE hFileToSend;
	cHTTPAction *phttpAction;
	

	static MyStack<SOCKET, MAX_SOCKET_POOL> stkSocketPool;
	static MyStack<cSocket*, MAX_OBJ_POOL> stkSocketObjects;


public:
	SOCKET socket;
	OverlappedEx olex;

	unsigned long len;
	char buffer[MAX_BUFFER_SIZE+1];


	static HANDLE hIOCompletionPort;
	static HANDLE hevDeath;
	static HANDLE hevContinue;

private:
	static DWORD WINAPI ListenThreadRoutine(void * pv);
	static DWORD WINAPI WorkerThreadRoutine(void * pv);


	void _init(void)
	{
		ATraceLn("cSocket::init(#o:%d, #s:%d)0x%lx", ::gSocketCounter, ::nSocketCount, this);
		memset(&this->olex, 0, sizeof(this->olex));
		this->olex.pSocket = this;
		this->olex.ets = etsInit;
		this->buffer[0] = 0;
		this->szCommand[0]=0;
		this->szResource[0]=0;
		this->socket=INVALID_SOCKET;
		this->len=0;
		this->nLenCommand=0;
		this->nLenResource=0;
		this->hFileToSend=INVALID_HANDLE_VALUE;
		this->eiohState=eioshBeginHeader;
		this->phttpAction = 0;
	}

	void _close(void)
	{
		ATraceLn("cSocket::close<%d>0x%lx", ::gSocketCounter, this);
		if (INVALID_HANDLE_VALUE != this->hFileToSend)
			::CloseHandle(this->hFileToSend);

		if (INVALID_SOCKET != this->socket)
		{
			cSocket::PushToPool(this->socket);
		}

		delete this->phttpAction;
	}

public:
	cSocket()
	{
		::InterlockedIncrement(&::gSocketCounter);
	}

	virtual ~cSocket()
	{
		::InterlockedDecrement(&::gSocketCounter);
	}

	void AbnormalDeath()
	{
		ATraceLn("cSocket::AbnormalDeath(0x%lx), s:0x%lx", this, this->socket);
		if (INVALID_SOCKET != this->socket)
		{
			::closesocket(this->socket);
			this->socket = INVALID_SOCKET;
			::InterlockedDecrement(&::nSocketCount);
			
			cSocket::PushToPool();
		}

		this->Close();
	}

	bool Init(SOCKET listener, OVERLAPPED *pov);

	void Close(void)
	{
		this->_close();
		cSocket::stkSocketObjects.Push(this);
	}

	bool PrimeIOPort(void)
	{

		this->olex.ets = etsRead;
		if (!::ReadFile((HANDLE) this->socket, this->buffer, sizeof(this->buffer)-1, NULL, (OVERLAPPED*)&this->olex))
		{
			DWORD err = ::GetLastError();

			return (ERROR_IO_PENDING != err);
		}

		return false;
	}

	bool DoRead(void)
	{
#ifdef _DEBUG
		this->buffer[this->len] = 0;
#endif
		for (unsigned int i = 0; i < this->len; ++i)
		{
			const char ch = this->buffer[i];
			switch (this->eiohState)
			{
				case eioshBeginHeader:
					if (isalpha(ch))
					{
						this->szCommand[this->nLenCommand++] = ch;
						this->eiohState = eioshReadCommand;
					}
				break;

				case eioshReadCommand:
					if (!isspace(ch))
					{
						this->szCommand[this->nLenCommand++] = ch;
					}
					else
					{
						this->szCommand[this->nLenCommand] = 0;
						this->eiohState = eioshReadResource;
					}
				break;

				case eioshReadResource:
					if (!isspace(ch))
					{
						this->szResource[this->nLenResource++] = ch;
					}
					else
					{
						this->szResource[this->nLenResource] = 0;
						this->eiohState = eioshBeginLine;
					}
				break;

				case eioshBeginLine:
					if ('\r' == ch)
						this->eiohState = eioshCR1;
				break;

				case eioshCR1:
					if ('\n' == ch)
						this->eiohState = eioshLF1;
					else
						this->eiohState = eioshBeginLine;
				break;

				case eioshLF1:
					if ('\r' == ch)
						this->eiohState = eioshCR2;
					else
						this->eiohState = eioshBeginLine;
				break;

				case eioshCR2:
					if ('\n' == ch)
						this->eiohState = eioshLF2;
					else
						this->eiohState = eioshBeginLine;
				break;
			}

		}

		if (eioshLF2 == this->eiohState)
		{
			ATraceLn("Command(%s), Rsrc(%s)", this->szCommand, this->szResource);
			this->DoWrite();
		}
		else
		{
			return this->PrimeIOPort();
		}

		return false;
	}	

	void DoTransmit(bool fFinal)
	{
		TRANSMIT_FILE_BUFFERS tfb={0};

		tfb.Head = this->buffer;
		tfb.HeadLength = this->len;
		this->olex.ets = fFinal? etsWriteAndDisconnect: etsWrite;
		if (::TransmitFile(this->socket, this->hFileToSend, 0, 0, (OVERLAPPED*)&this->olex, &tfb, fFinal?(TF_REUSE_SOCKET|TF_DISCONNECT): 0))
		{
			ATraceLn("Synchronous transmission?");
		}
		else
		{
			DWORD err = ::GetLastError();

			if (ERROR_IO_PENDING != err)
			{
				ATraceLn("Error Transmitting:%d0x%x", err, err);
				this->AbnormalDeath();
			}
		}

	}


	void DoWrite(void)
	{
		char *pszAction = strchr(&this->szResource[1], '?');
		if (!pszAction)
		{
			this->hFileToSend = ::CreateFile(&this->szResource[1], GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
			if (INVALID_HANDLE_VALUE == this->hFileToSend)
			{
				ATraceLn("Error opening file(0x%x)", ::GetLastError());
				this->len = wsprintf(this->buffer, "HTTP/1.0 404 \r\nContent-Type: text/html\r\n\r\n");
			}
			else
			{
				this->len = wsprintf(this->buffer, "HTTP/1.0 200 \r\nContent-Type: text/html\r\n\r\n");
			}
		}
		else
		{
			this->hFileToSend = 0;

			pszAction = strtok(&this->szResource[1], "?");

			if (0 == strcmpi("debug", pszAction))
			{
				static int i=0;

				this->len = wsprintf(this->buffer, "HTTP/1.0 200 \r\nContent-Type: text/html\r\n\r\nHello %d(#o:%d, #s:%d)\n", i++, ::gSocketCounter, ::nSocketCount);

			}
			else if (0 == strcmpi("hello", pszAction))
			{
				this->len = wsprintf(this->buffer, "HTTP/1.0 200 \r\nContent-Type: text/html\r\n\r\n");

				char *pszValue = strtok(0, "=");
				if (0 == strcmpi("numlines", pszValue))
				{
					pszValue = strtok(0, ",");
					if (pszValue)
					{
						int num = atoi(pszValue);
						
						for (int j =0; j< num; ++j)
						{
							const char szHello[]="Hello ";
							const char szBR[]="<br>";

							strcpy(&this->buffer[this->len], szHello);
							this->len += sizeof(szHello)-1;
							this->len += myitoa(j, &this->buffer[this->len]);
							strcpy(&this->buffer[this->len], szBR);
							this->len += sizeof(szBR)-1;


							if (this->len > sizeof(this->buffer)*9/10)
								break;
						}
					}
				}
			}
			else
			{
				ATraceLn("Unknown action:%s", pszAction);
				this->len = wsprintf(this->buffer, "HTTP/1.0 404 \r\n\r\n");
			}
		}

		this->DoTransmit(true);

	}

	void ProcessIO()
	{
		bool fRead = (etsRead == this->olex.ets);

		ATraceLn("\tProcessIO(%s, %lx):0x%lx, len:%d", fRead?"r":"w", this->socket, this, this->len);
		if (0== this->len)
		{
			this->AbnormalDeath();
		}
		else if (fRead)
		{
			if (this->DoRead())
				this->AbnormalDeath();
		}
		else if (etsWriteAndDisconnect == this->olex.ets)
		{
			this->Close();
		}
	}


	static bool InitSockets(void);
	static void PumpSockets(long nThreads);
	static void CloseSockets()
	{
		::WSACleanup();
	}

	static void PushToPool(void)
	{
		SOCKET socket;

		socket = ::socket(AF_INET, SOCK_STREAM, 0);
		if (INVALID_SOCKET == socket)
		{
				ATraceLn("Error: could not create socket:0x%lx", ::GetLastError());
		}
		else
		{
			if (!::CreateIoCompletionPort((HANDLE)socket, cSocket::hIOCompletionPort, 0, 0))
			{
				ATraceLn("Error: could not create io completion port:0x%lx", ::GetLastError());
				::closesocket(socket);
			}
			else
			{
				if (!cSocket::stkSocketPool.Push(socket))
				{
					::InterlockedIncrement(&::nSocketCount);
				}

			}
		}

	}

	static void PushToPool(SOCKET socket)
	{
		if (cSocket::stkSocketPool.Push(socket))
		{
			::closesocket(socket);
			::InterlockedDecrement(&::nSocketCount);
		}
	}

	static SOCKET PopFromPool(void)
	{
		SOCKET socket = INVALID_SOCKET;

		if (cSocket::stkSocketPool.Pop(&socket))
		{
			for (int i = 0; i <= 4; ++i)
			{
				cSocket::PushToPool();
			}

			cSocket::stkSocketPool.Pop(&socket);
		}
		return socket;
	}
};


HANDLE cSocket::hIOCompletionPort = NULL;
HANDLE cSocket::hevDeath = NULL;
HANDLE cSocket::hevContinue = NULL;

MyStack<SOCKET, MAX_SOCKET_POOL> cSocket::stkSocketPool;
MyStack<cSocket*, MAX_OBJ_POOL> cSocket::stkSocketObjects;


DWORD WINAPI cSocket::WorkerThreadRoutine(void * pv)
{
	while(!cSocket::hIOCompletionPort)
		Sleep(1000);

	while (true)
	{
		BOOL fok;
		DWORD nBytes=0;
		DWORD dummy;
		cSocket *pSock=0;
		cSocket::OverlappedEx *polex=0;

		fok = ::GetQueuedCompletionStatus(cSocket::hIOCompletionPort, &nBytes, &dummy, (OVERLAPPED**)&polex, INFINITE);
		pSock = polex->pSocket;
		if (fok && pSock)
		{
			pSock->len = nBytes;
			pSock->ProcessIO();
		}
		else
		{
			DWORD err = ::GetLastError();
			ATraceLn("Err:%ld,0x%lx", err, err);
			if (pSock)
				pSock->AbnormalDeath();
		}
	}

	return 0;
}


bool cSocket::
Init(SOCKET listener, OVERLAPPED *pov)
{
	this->_init();

	this->socket = cSocket::PopFromPool();

	if (INVALID_SOCKET == this->socket)
		return true;


	DWORD err;

	ATraceLn("cSocket::Init (0x%lx), s:%lx", this, this->socket);
	if (::AcceptEx(listener, this->socket, this->buffer, 0,sizeof(SOCKADDR_IN)+16,sizeof(SOCKADDR_IN)+16, &this->len, pov))
	{
		ATraceLn("Syncronous AcceptEx?");
	}
	else
	{
		err = ::WSAGetLastError();
		if (ERROR_IO_PENDING != err)
		{
			ATraceLn("AcceptEx err:0x%lx", err);
			return true;
		}
	}

	return false;
}

DWORD WINAPI cSocket::ListenThreadRoutine(void * pv)
{
	int nThreads = (int)pv;
	SOCKET listener;
	SOCKADDR_IN localaddr={0};

	ATraceLn("Creating Listener...");
	listener = ::socket(AF_INET, SOCK_STREAM, 0);
	if (INVALID_SOCKET == listener)
	{
		ATraceLn("Error:%d", WSAGetLastError());
		return 1;
	}

	localaddr.sin_port = htons(::nPortNum);
	localaddr.sin_family = AF_INET;

	ATraceLn("Binding Listener...");
	DWORD res;
	do
	{
		res = bind(listener, (sockaddr*)&localaddr, sizeof(localaddr));
		if (SOCKET_ERROR == res)
		{
			ATraceLn("Error:%d", WSAGetLastError());
			::Sleep(100);
		}
	}
	while(SOCKET_ERROR == res);


	::setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, (char*)TRUE, sizeof(int));

	ATraceLn("Start Listening...");
	res = listen(listener, SOMAXCONN);
	if (SOCKET_ERROR == res)
	{
		ATraceLn("Error:%d", WSAGetLastError());
		return 1;
	}

	HANDLE rghEvents[2] = {0};
	OVERLAPPED ov = {0};

	ov.hEvent= rghEvents[0] = ::CreateEvent(NULL, FALSE, FALSE, NULL);
	rghEvents[1] = cSocket::hevDeath;

	ATraceLn("Enter Acceptance loop...");
	while (true)
	{
		cSocket * pSock=0;

		ATraceLn("objtop:%d, socktop:%d", cSocket::stkSocketObjects.Top(), cSocket::stkSocketPool.Top());
		if (cSocket::stkSocketObjects.Pop(&pSock))
			::Sleep(10);
		else if (pSock)
		{
			if (pSock->Init(listener, &ov))
				pSock->AbnormalDeath();
			else
			{
				DWORD dwWaitResult;


				dwWaitResult = ::WaitForMultipleObjects(sizeof(rghEvents)/sizeof(HANDLE), rghEvents, FALSE, INFINITE);
				if(WAIT_OBJECT_0 != dwWaitResult || pSock->PrimeIOPort())
					pSock->AbnormalDeath();

				::ResetEvent(ov.hEvent);

				if (WAIT_OBJECT_0 + 1 == dwWaitResult)
					break;
			}
		}
	}

	::closesocket(listener);
	ATraceLn("Done Listening");
	::Sleep(5*1000);
	::SetEvent(cSocket::hevContinue);

	return 0;
}


bool cSocket::InitSockets()
{
	WSADATA wsaData;

	if (!::WSAStartup(MAKEWORD(2,0), &wsaData))
	{
		cSocket::hIOCompletionPort = ::CreateIoCompletionPort(INVALID_HANDLE_VALUE, 0, 0, 0);
		if (!cSocket::hIOCompletionPort)
			return true;

		cSocket::hevDeath = ::CreateEvent(NULL, FALSE, FALSE, "Rani Death Event");
		if (!cSocket::hevDeath)
			return true;
		
		cSocket::hevContinue = ::CreateEvent(NULL, FALSE, FALSE, "Rani Continue Event");
		if (!cSocket::hevContinue)
			return true;

		if (ERROR_ALREADY_EXISTS == ::GetLastError())
		{
			::ResetEvent(cSocket::hevContinue);
			::SetEvent(cSocket::hevDeath);

			ATraceLn("Waiting for other instance to die...");
			::WaitForSingleObject(cSocket::hevContinue, INFINITE);
			ATraceLn("Done waiting!");
		}

		for (int i = 0; i < MAX_SOCKET_POOL; ++i)
		{
			cSocket::PushToPool();
		}

		for (i = 0; i<MAX_OBJ_POOL; ++i)
		{
			cSocket::stkSocketObjects.Push(new cSocket);
		}
	}

	return false;
}

void cSocket::PumpSockets(long nThreads)
{
	HANDLE hThread;
	DWORD idThread;

	hThread = ::CreateThread(0, 0, cSocket::ListenThreadRoutine, (void *)nThreads,0, &idThread);
	if (hThread)
	{
		HANDLE h;
		while(nThreads--)
		{
			DWORD idThread;

			h=CreateThread(0, 0, cSocket::WorkerThreadRoutine, (void *)nThreads,0, &idThread);
		}

		::WaitForSingleObject(h, INFINITE);
		::WaitForSingleObject(hThread, INFINITE);
	}
}


void main(int argc, char ** argv)
{
	enum
	{
		stParam,
		stReadThreadCount,
		stReadPortnumber
	} state = stParam;

	int nWorkThreads = 1;

	for (int i = 1; i < argc; i++)
	{
		char *psz = argv[i];

		switch(state)
		{
			case stParam:
				switch(*psz)
				{
					case '/':
					case '-':
						switch(tolower(*(++psz)))
						{
							case 't':
								state = stReadThreadCount;
							break;

							case 'p':
								state = stReadPortnumber;
							break;
						}
					break;
				}
			break;

			case stReadThreadCount:
				state = stParam;
				nWorkThreads = atoi(argv[i]);
				nWorkThreads = min(nWorkThreads, 100000);
			break;

			case stReadPortnumber:
				state = stParam;
				::nPortNum = atoi(argv[i]);
			break;
		}
	}

	if (!cSocket::InitSockets())
	{
		cSocket::PumpSockets(nWorkThreads);
		cSocket::CloseSockets();
	}
}
